from typing import Type, Tuple
from pathlib import Path
import shutil
import click
from textual.design import ColorSystem
from pydantic import BaseModel, Field
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
)

app_dir = Path(click.get_app_dir(app_name="lazyfeed"))
config_file_path = app_dir / "config.toml"
template_file_path = Path(__file__).parent / "config_template.toml"


class ClientSettings(BaseModel):
    timeout: int = 300
    connect_timeout: int = 10
    headers: dict = {}


class Theme(BaseModel):
    primary: str | None = Field(default="#7dc4e4")
    secondary: str | None = Field(default="#cad3f5")
    background: str | None = Field(default="#24273a")
    surface: str | None = Field(default="#5b6078")
    success: str | None = Field(default="#a6da95")
    warning: str | None = Field(default="#f5a97f")
    error: str | None = Field(default="#ed8796")
    dark: bool = True

    def to_color_system(self) -> ColorSystem:
        """
        Convert current theme to a ColorSystem.
        """

        return ColorSystem(**self.model_dump())


class AppSettings(BaseModel):
    db_url: str = f"sqlite:///{app_dir / 'lazyfeed.db'}"

    theme: Theme = Field(default_factory=Theme)

    auto_mark_as_read: bool = False
    ask_before_marking_as_read: bool = False
    show_read: bool = False
    sort_by: str = "title"
    sort_order: str = "descending"


class Settings(BaseSettings):
    client: ClientSettings = Field(default_factory=ClientSettings)
    app: AppSettings = Field(default_factory=AppSettings)

    model_config = SettingsConfigDict(
        env_nested_delimiter="__",
        toml_file=f"{app_dir / 'config.toml'}",
        validate_default=True,
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        app_dir.mkdir(parents=True, exist_ok=True)

        if not config_file_path.exists():
            shutil.copy(template_file_path, config_file_path)

        return (
            env_settings,
            TomlConfigSettingsSource(settings_cls),
        )


if __name__ == "__main__":
    settings = Settings()
    print(settings.model_dump())

    print(settings.app.theme.to_color_system().generate())



import xml.etree.ElementTree as ET
from lazyfeed.models import Feed


def export_opml(feeds: list[Feed], output_file):
    opml = ET.Element("opml", version="1.0")

    head = ET.SubElement(opml, "head")
    title = ET.SubElement(head, "title")
    title.text = "RSS feeds from lazyfeed"
    body = ET.SubElement(opml, "body")

    for feed in feeds:
        ET.SubElement(body, "outline", text=feed.title, type="rss", xmlUrl=feed.url)

    tree = ET.ElementTree(opml)
    tree.write(output_file, encoding="utf-8", xml_declaration=True)


def import_opml(input_file) -> list[str]:
    tree = ET.parse(input_file)
    root = tree.getroot()

    feeds = []
    for outline in root.findall(".//outline"):
        xml_url = outline.get("xmlUrl")
        if xml_url:
            feeds.append(xml_url)

    return feeds



from aiohttp import web
import pytest
from lazyfeed.feeds import fetch_feed_metadata, fetch_feed, fetch_post
from lazyfeed.models import Feed

feed = """<?xml version="1.0" encoding="utf-8"?>
<rss version="2.0" xmlns:atom="http://www.w3.org/2005/Atom">
    <channel>
        <title>Example RSS</title>
        <link>https://example.com/</link>
        <description>Latest posts from example's blog</description>
        <atom:link href="https://example.com/rss/" rel="self"/>
        <lastBuildDate>Thu, 15 Aug 2024 19:55:29 +0000</lastBuildDate>
        <item>
            <title>Example post</title>
            <link>https://example.com/blog/post/</link>
            <description><![CDATA[A simple change than can protect you and your site from basic brute force attacks.]]></description>
            <pubDate>Thu, 15 Aug 2024 19:50:00 +0000</pubDate>
            <guid>https://example.com/blog/post/</guid>
        </item>
    </channel>
</rss>
"""

empty_feed = """<?xml version="1.0" encoding="utf-8"?>
<rss version="2.0">
    <channel>
        <title>Empty RSS</title>
        <link>https://example.com/</link>
        <description>No posts available.</description>
    </channel>
</rss>
"""

post = """<html>
<head>
<title>Post</title>
</head>
<body>
<h1>Example post</h1>
<p>Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</p>
</body>
</html>
"""


async def rss_success(request):
    return web.Response(status=200, text=feed)


async def rss_bad_format(request):
    return web.Response(status=200, text="\n" + feed)


async def rss_empty(request):
    return web.Response(status=200, text="")


async def rss_not_found(request):
    return web.Response(status=404, text="")


async def rss_not_modified(request):
    return web.Response(status=304, text="")


async def post_success(request):
    return web.Response(status=200, text=post)


async def post_not_found(request):
    return web.Response(status=404, text="")


def create_app():
    app = web.Application()
    app.router.add_get("/rss_success", rss_success)
    app.router.add_get("/rss_bad_format", rss_bad_format)
    app.router.add_get("/rss_empty", rss_empty)
    app.router.add_get("/rss_not_found", rss_not_found)
    app.router.add_get("/rss_not_modified", rss_not_modified)
    app.router.add_get("/post_success", post_success)
    app.router.add_get("/post_not_found", post_not_found)

    return app


async def test_fetch_feed_bad_url(aiohttp_client):
    client = await aiohttp_client(create_app())
    with pytest.raises(RuntimeError):
        await fetch_feed_metadata(client, "/")


async def test_fetch_feed_metadata_success(aiohttp_client):
    client = await aiohttp_client(create_app())
    result = await fetch_feed_metadata(client, "/rss_success")

    assert result is not None
    assert result.title == "Example RSS"
    assert result.link == "https://example.com/"


async def test_fetch_feed_metadata_bad_format(aiohttp_client):
    client = await aiohttp_client(create_app())
    with pytest.raises(RuntimeError):
        await fetch_feed_metadata(client, "/rss_bad_format")


async def test_fetch_feed_metadata_not_found(aiohttp_client):
    client = await aiohttp_client(create_app())
    with pytest.raises(RuntimeError):
        await fetch_feed_metadata(client, "/rss_not_found")


async def test_fetch_feed_success(aiohttp_client):
    feed = Feed(
        url="/rss_success",
        link="https://example.com",
        title="Example",
    )
    client = await aiohttp_client(create_app())
    entries, etag = await fetch_feed(client, feed)
    assert len(entries) == 1
    assert etag == ""


async def test_fetch_feed_success_but_empty(aiohttp_client):
    feed = Feed(
        url="/rss_empty",
        link="https://example.com",
        title="Example",
    )
    client = await aiohttp_client(create_app())
    entries, etag = await fetch_feed(client, feed)
    assert len(entries) == 0
    assert etag == ""


async def test_fetch_feed_not_modified(aiohttp_client):
    feed = Feed(
        url="/rss_not_modified",
        link="https://example.com",
        title="Example",
        etag="12345",
    )

    client = await aiohttp_client(create_app())
    entries, etag = await fetch_feed(client, feed)
    assert len(entries) == 0
    assert etag == "12345"


async def test_fetch_post_bad_url(aiohttp_client):
    client = await aiohttp_client(create_app())
    with pytest.raises(RuntimeError):
        await fetch_post(client, "/")


async def test_fetch_post_success(aiohttp_client):
    client = await aiohttp_client(create_app())
    content = await fetch_post(client, "/post_success")
    assert content is not None


async def test_fetch_post_not_found(aiohttp_client):
    client = await aiohttp_client(create_app())
    with pytest.raises(RuntimeError):
        await fetch_post(client, "/post_not_found")



import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from lazyfeed.repositories import FeedRepository, PostRepository
from lazyfeed.models import Base, Feed


@pytest.fixture(scope="function")
def session():
    engine = create_engine("sqlite:///:memory:")
    Base.metadata.create_all(engine)

    with Session(engine) as session:
        yield session


@pytest.fixture
def feed_repository(session):
    return FeedRepository(session)


@pytest.fixture
def posts_repository(session):
    return PostRepository(session)


class TestFeedRepository:
    @pytest.fixture(autouse=True)
    def setup(self, feed_repository):
        self.feed_repository = feed_repository
        self.feed = Feed(
            url="https://dontpanic.com/feed",
            title="The Infinite Improbability Newsfeed",
        )

    def test_add(self):
        assert self.feed.id is None

        feed_in_db = self.feed_repository.add(self.feed)
        assert feed_in_db.id is not None
        assert feed_in_db.created_at is not None
        assert feed_in_db.last_updated_at is not None
        assert feed_in_db.id == 1
        assert feed_in_db.title == self.feed.title
        assert feed_in_db.url == self.feed.url

    def test_get(self):
        feed_in_db = self.feed_repository.get(1)
        assert feed_in_db is None

        self.feed_repository.add(self.feed)

        feed_in_db = self.feed_repository.get(1)
        assert feed_in_db is not None
        assert feed_in_db.url == self.feed.url
        assert feed_in_db.title == self.feed.title

    def test_get_by_attributes(self):
        feeds_in_db = self.feed_repository.get_by_attributes(title=self.feed.title)
        assert len(feeds_in_db) == 0

        self.feed_repository.add(self.feed)

        feeds_in_db = self.feed_repository.get_by_attributes(title=self.feed.title)
        assert len(feeds_in_db) == 1
        assert feeds_in_db[0].title == self.feed.title

    def test_get_all(self):
        for i in range(5):
            self.feed_repository.add(
                Feed(
                    title=f"https://news.{ i + 1}.com/rss",
                    url=f"News {i + 1}",
                )
            )

        feeds_in_db = self.feed_repository.get_all()
        assert len(feeds_in_db) == 5

    def test_update(self):
        original_feed = self.feed_repository.add(self.feed)
        original_feed_title = self.feed.title
        self.feed_repository.update(
            original_feed.id,
            title="The Finite Certainty Bulletin",
        )

        updated_feed = self.feed_repository.get(original_feed.id)
        assert updated_feed is not None
        assert updated_feed.title != original_feed_title

    def test_delete(self):
        feed_in_db = self.feed_repository.add(self.feed)
        assert feed_in_db is not None

        self.feed_repository.delete(feed_in_db.id)
        deleted_feed = self.feed_repository.get(feed_in_db.id)
        assert deleted_feed is None



from pathlib import Path
from textual.app import ComposeResult
from textual.containers import VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import Markdown


class HelpModal(ModalScreen):
    BINDINGS = [
        ("escape", "app.pop_screen", "Quit"),
        ("q", "app.pop_screen", "Quit"),
        ("?", "app.pop_screen", "Quit"),
    ]

    def compose(self) -> ComposeResult:
        markdown_path = Path(__file__).parent / "help_modal.md"
        with open(markdown_path, "r") as f:
            markdown = f.read()

        with VerticalScroll(classes="modal modal--help"):
            yield Markdown(markdown=markdown)

    def on_mount(self) -> None:
        self.modal = self.query_one(VerticalScroll)
        self.modal.border_title = "help"
        self.modal.border_subtitle = "q/? quit"
        self.modal.focus()



import asyncio
import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from lazyfeed.db import init_db
from lazyfeed.models import Feed, Post
from lazyfeed.repositories import FeedRepository, PostRepository
from lazyfeed.settings import Settings
from lazyfeed.tui import ActiveView, LazyFeedApp


@pytest.fixture(scope="function")
def settings():
    settings = Settings()
    settings.app.db_url = "sqlite:///:memory:"
    return settings


@pytest.fixture(scope="function")
def session(settings):
    engine = create_engine(settings.app.db_url)
    init_db(engine)
    session = Session(engine)

    yield session

    session.close()


def _populate_db(session) -> int:
    feed_repo = FeedRepository(session)
    feed = feed_repo.add(
        Feed(
            url="http://hitchhikers.com/rss",
            title="The Hitchhiker's Guide to the Galaxy",
            link="http://hitchhikers.com",
            description="A guide to the galaxy, filled with useful information and humor.",
        )
    )

    post_repo = PostRepository(session)
    posts = [
        Post(
            feed_id=feed.id,
            url="http://hitchhikers.com/item1",
            author="Douglas Adams",
            title="Don't Panic",
            summary="The most important phrase in the universe.",
            content="Always know where your towel is.",
        ),
        Post(
            feed_id=feed.id,
            url="http://hitchhikers.com/item2",
            author="Douglas Adams",
            title="The Answer to Life, the Universe, and Everything",
            summary="The number 42.",
            content="The answer to the ultimate question of life, the universe, and everything is 42.",
        ),
        Post(
            feed_id=feed.id,
            url="http://hitchhikers.com/item3",
            author="Douglas Adams",
            title="The Hitchhiker's Guide",
            summary="A book that tells you everything you need to know.",
            content="A guide that provides information about the galaxy.",
        ),
        Post(
            feed_id=feed.id,
            url="http://hitchhikers.com/item4",
            author="Douglas Adams",
            title="The Infinite Improbability Drive",
            summary="A drive that allows for improbable events.",
            content="A drive that makes the impossible possible.",
        ),
    ]

    for post in posts:
        post_repo.add(post)

    return len(posts)


async def test_views(settings, session):
    app = LazyFeedApp(session, settings)

    async with app.run_test() as pilot:
        assert app.tabloid.border_title == "lazyfeed"
        assert app.active_view == ActiveView.IDLE

        await pilot.press("g", "l")
        assert app.tabloid.border_title == "lazyfeed/saved"
        assert app.active_view == ActiveView.SAVED

        await pilot.press("g", "a")
        assert app.tabloid.border_title == "lazyfeed/all"
        assert app.active_view == ActiveView.ALL

        await pilot.press("g", "f")
        assert app.tabloid.border_title == "lazyfeed/fav"
        assert app.active_view == ActiveView.FAV

        await pilot.press("g", "p")
        assert app.tabloid.border_title == "lazyfeed"
        assert app.active_view == ActiveView.PENDING


async def test_mark_item_as_read(settings, session):
    num_posts = _populate_db(session)

    settings.app.show_read = False
    app = LazyFeedApp(session, settings)

    async with app.run_test() as pilot:
        await asyncio.sleep(1)
        assert app.tabloid.row_count == num_posts

        await pilot.press("m")
        assert app.tabloid.row_count == num_posts - 1


async def test_mark_all_items_as_read(settings, session):
    num_posts = _populate_db(session)

    settings.app.show_read = False
    app = LazyFeedApp(session, settings)

    async with app.run_test() as pilot:
        await asyncio.sleep(1)
        assert app.tabloid.row_count == num_posts

        await pilot.press("A")
        assert app.tabloid.row_count == 0



import asyncio
import os
import subprocess
from rich.console import Console
from rich.table import Table
from sqids import Sqids
from sqlalchemy import create_engine, exc, text
from sqlalchemy.orm import Session
import aiohttp
import rich_click as click
from lazyfeed.settings import Settings, config_file_path
from lazyfeed.db import init_db
from lazyfeed.feeds import fetch_feed_metadata
from lazyfeed.models import Feed
from lazyfeed.opml_utils import export_opml, import_opml
from lazyfeed.repositories import FeedRepository, PostRepository
from lazyfeed.tui import LazyFeedApp


console = Console()
sqids = Sqids(
    alphabet="e69auyponmz7lk5vtgwi1r23hb0d8jq4xfcs",
    min_length=3,
)


@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx) -> None:
    """
    A modern RSS feed reader for the terminal.
    """

    ctx.ensure_object(dict)

    # Load settings
    settings = Settings()
    ctx.obj["settings"] = settings

    # Set up the SQLite database engine.
    engine = create_engine(settings.app.db_url)
    init_db(engine)

    ctx.obj["engine"] = engine

    # If no subcommand, start the TUI.
    if ctx.invoked_subcommand is None:
        ctx.forward(start_tui)


@cli.command(
    name="tui",
    help="Start the TUI.",
)
@click.pass_context
def start_tui(ctx) -> None:
    engine = ctx.obj["engine"]
    settings = ctx.obj["settings"]
    with Session(engine) as session:
        app = LazyFeedApp(session, settings)
        app.run()


async def _add_feeds(session: Session, settings: Settings, urls: list[str]):
    feed_repository = FeedRepository(session)
    already_saved_urls = [feed.url for feed in feed_repository.get_all()]
    new_urls = [url for url in urls if url not in already_saved_urls]

    if not new_urls:
        console.print("[red]ERR[/] There are no new urls to check!")
        return
    with console.status(
        "Fetching new feeds... This will only take a moment!",
        spinner="earth",
    ):
        timeout = aiohttp.ClientTimeout(
            total=settings.client.timeout,
            connect=settings.client.connect_timeout,
        )
        headers = settings.client.headers
        async with aiohttp.ClientSession(timeout=timeout, headers=headers) as client:
            tasks = [fetch_feed_metadata(client, url) for url in urls]
            results = await asyncio.gather(*tasks, return_exceptions=True)
            for url, result in zip(urls, results):
                if isinstance(result, Exception):
                    console.print(f"[red]ERR[/] [link={url}]{url}[/]")
                else:
                    assert isinstance(result, Feed)

                    feed_in_db = feed_repository.add(result)
                    console.print(
                        f"[green]OK[/] [link={feed_in_db.link}]{feed_in_db.title}[/]"
                    )


@cli.command(
    name="add",
    help="Add one or more RSS feeds.",
)
@click.argument("urls", nargs=-1)
@click.pass_context
def add_feed(ctx, urls) -> None:
    engine = ctx.obj["engine"]
    settings = ctx.obj["settings"]
    with Session(engine) as session:
        asyncio.run(_add_feeds(session, settings, urls))


@cli.command(
    name="list",
    help="Print a list with all your RSS feeds.",
)
@click.pass_context
def list_feeds(ctx):
    engine = ctx.obj["engine"]
    with Session(engine) as session:
        feed_repository = FeedRepository(session)
        feeds = feed_repository.get_all()

        if not len(feeds):
            console.print("[red]ERR[/] Add some feeds first!")
            return

        table = Table(
            show_header=True,
            show_lines=True,
        )
        table.add_column("id", justify="center")
        table.add_column("title", justify="left")

        for feed in feeds:
            table.add_row(
                f"[bold]{sqids.encode([feed.id])}[/]",
                f"[link={feed.link}]{feed.title}[/]",
            )

        console.print(table)
        console.print(
            f"{len(feeds)} feeds",
            highlight=False,
            justify="center",
        )


@cli.command(
    name="delete",
    help="Remove specified RSS feed.",
)
@click.argument("feed_id")
@click.pass_context
def delete_feed(ctx, feed_id):
    engine = ctx.obj["engine"]
    with Session(engine) as session:
        feed_repo = FeedRepository(session)
        post_repo = PostRepository(session)

        decoded_id = sqids.decode(feed_id)[0]
        feed = feed_repo.delete(decoded_id)
        if not feed:
            console.print(f"[red]ERR[/] No feed found with ID '{feed_id}'.")
            return

        posts = post_repo.get_by_attributes(feed_id=feed.id)
        for post in posts:
            post_repo.delete(post.id)

        console.print(
            f"[green]OK[/] [link={feed.url}]{feed.title}[/] deleted correctly!"
        )


@cli.command(
    name="import",
    help="Import RSS feeds from an OPML file.",
)
@click.argument("input", type=click.File("rb"))
@click.pass_context
def import_feeds(ctx, input) -> None:
    engine = ctx.obj["engine"]
    settings = ctx.obj["settings"]
    with Session(engine) as session:
        urls = import_opml(input)
        asyncio.run(_add_feeds(session, settings, urls))


@cli.command(
    name="export",
    help="Export all RSS feeds to an OPML file.",
)
@click.argument(
    "output",
    type=click.File("wb"),
    default="lazyfeed.opml",
)
@click.pass_context
def export_feeds(ctx, output) -> None:
    engine = ctx.obj["engine"]
    with Session(engine) as session:
        feed_repository = FeedRepository(session)
        feeds = feed_repository.get_all()

        if not len(feeds):
            console.print("[red]ERR[/] Add some feeds first!")
            return

        export_opml(feeds, output)


@cli.command(name="config", help="Open the configuration file.")
def config() -> None:
    user_editor = os.environ.get("EDITOR", None)
    try:
        if user_editor:
            subprocess.run([user_editor, str(config_file_path)], check=True)
        else:
            subprocess.run(["open", str(config_file_path)], check=True)
    except Exception as e:
        console.print(f"[red]ERR[/] Failed to open the configuration file: {e}")


@cli.command(
    name="vacuum",
    help="Reclaim unused spaced in the database.",
)
@click.pass_context
def vacuum(ctx) -> None:
    engine = ctx.obj["engine"]
    with Session(engine) as session:
        try:
            session.execute(text("VACUUM"))
            session.commit()
            console.print("[green]OK[/] Unused space reclaimed!")
        except exc.SQLAlchemyError as e:
            console.print(f"[red]ERR[/] Something went wrong: {e}!")



from sqlalchemy import select, update
from sqlalchemy.orm import Session
from lazyfeed.models import Feed, Post


class Repository[T: (Feed, Post)]:
    def __init__(self, session: Session, model) -> None:
        self.session = session
        self.model = model

    def add(self, entity: T) -> T:
        self.session.add(entity)
        self.session.commit()
        self.session.refresh(entity)

        return entity

    def add_in_batch(self, entities: list[T]) -> None:
        self.session.add_all(entities)
        self.session.commit()

    def get(self, id: int) -> T | None:
        stmt = select(self.model).where(self.model.id == id)
        return self.session.scalars(stmt).one_or_none()

    def get_by_attributes(self, **kwargs) -> list[T]:
        return self.session.query(self.model).filter_by(**kwargs).all()

    def get_all(self) -> list[T]:
        return self.session.query(self.model).all()

    def update(self, id: int, **kwargs) -> None:
        stmt = update(self.model).where(self.model.id == id).values(**kwargs)
        self.session.execute(stmt)
        self.session.commit()

    def delete(self, id: int) -> None:
        stmt = select(self.model).where(self.model.id == id)
        entity = self.session.scalars(stmt).one_or_none()
        if entity:
            self.session.delete(entity)
            self.session.commit()
            return entity


class FeedRepository(Repository[Feed]):
    def __init__(self, session: Session) -> None:
        super().__init__(session, Feed)


class PostRepository(Repository[Post]):
    def __init__(self, session: Session) -> None:
        super().__init__(session, Post)

    def get_sorted(self, sort_by: str, ascending: bool, **kwargs) -> list[Post]:
        sort_mapping = {
            "title": Post.title,
            "published_date": Post.published_at,
            "read_status": Post.read,
        }

        sort_criteria = sort_mapping.get(sort_by, Post.published_at)
        sort_order = sort_criteria.asc() if ascending else sort_criteria.desc()

        return self.session.query(Post).order_by(sort_order).filter_by(**kwargs).all()

    def mark_all_as_read(self) -> None:
        stmt = update(Post).where(Post.read == False).values(read=True)
        self.session.execute(stmt)
        self.session.commit()



import time
from textual import events
from textual.binding import Binding
from textual.message import Message
from textual.reactive import reactive
from textual.widgets import DataTable


class Tabloid(DataTable):
    def __init__(self, *args, **kwargs):
        super().__init__(cursor_type="row", header_height=0, *args, **kwargs)

    BINDINGS = DataTable.BINDINGS + [
        Binding("A", "mark_all_as_read", "Mark All As Read", show=False),
        Binding("G", "scroll_bottom", "Bottom", show=False),
        Binding("j", "cursor_down", "Cursor Down", show=False),
        Binding("k", "cursor_up", "Cursor Up", show=False),
        Binding("n", "cursor_down", "Cursor Down", show=False),
        Binding("p", "cursor_up", "Cursor Up", show=False),
        Binding("o", "select_cursor", "Open In Browser", show=False),
        Binding("m", "mark_as_read", "Mark Post As Read", show=False),
    ]

    first_key_pressed: reactive[str | None] = reactive(None)

    def on_mount(self) -> None:
        self.add_column("s", key="saved")
        self.add_column("f", key="fav")
        self.add_column("t", key="title")

        self.border_title = self.app.TITLE
        self.border_subtitle = "↑/k up · ↓/j down · o open · q quit · ? help"

    def action_select_cursor(self) -> None:
        try:
            row_key, _ = self.coordinate_to_cell_key(self.cursor_coordinate)
            self.post_message(self.OpenPost(int(row_key.value)))
        except Exception:
            pass

    def action_mark_as_read(self) -> None:
        try:
            row_key, _ = self.coordinate_to_cell_key(self.cursor_coordinate)
            self.post_message(self.MarkPostAsRead(int(row_key.value)))
        except Exception:
            pass

    def action_mark_all_as_read(self) -> None:
        self.post_message(self.MarkAllPostsAsRead())

    def action_save_for_later(self) -> None:
        try:
            row_key, _ = self.coordinate_to_cell_key(self.cursor_coordinate)
            self.post_message(self.SavePost(int(row_key.value)))
        except Exception:
            pass

    def action_mark_as_fav(self) -> None:
        try:
            row_key, _ = self.coordinate_to_cell_key(self.cursor_coordinate)
            self.post_message(self.MarkPostAsFav(int(row_key.value)))
        except Exception:
            pass

    async def on_key(self, event: events.Key) -> None:
        if event.key == "g":
            if self.first_key_pressed == "g":
                self.action_scroll_top()
                self.first_key_pressed = None
            else:
                self.first_key_pressed = "g"
                self.first_key_time = time.time()
        elif self.first_key_pressed == "g":
            if time.time() - self.first_key_time < 0.5:
                action_map = {
                    "l": self.LoadSavedPosts,
                    "f": self.LoadFavPosts,
                    "p": self.LoadAllNewPosts,
                    "n": self.LoadAllNewPosts,
                    "a": self.LoadAllPosts,
                }
                action = action_map.get(event.key)
                if action:
                    self.post_message(action())

            self.first_key_pressed = None
        elif event.key == "f":
            self.action_mark_as_fav()
            self.first_key_pressed = None
        elif event.key == "s":
            self.action_save_for_later()
            self.first_key_pressed = None
        else:
            self.first_key_pressed = None

    class LoadAllPosts(Message):
        pass

    class LoadAllNewPosts(Message):
        pass

    class LoadSavedPosts(Message):
        pass

    class LoadFavPosts(Message):
        pass

    class OpenPost(Message):
        def __init__(self, post_id: int) -> None:
            super().__init__()
            self.post_id = post_id

    class MarkPostAsRead(Message):
        def __init__(self, post_id: int) -> None:
            super().__init__()
            self.post_id = post_id

    class MarkAllPostsAsRead(Message):
        pass

    class SavePost(Message):
        def __init__(self, post_id: int) -> None:
            super().__init__()
            self.post_id = post_id

    class MarkPostAsFav(Message):
        def __init__(self, post_id: int) -> None:
            super().__init__()
            self.post_id = post_id



from typing import Any
import aiohttp
import feedparser
from lazyfeed.models import Feed


async def fetch_feed_metadata(client: aiohttp.ClientSession, feed_url: str) -> Feed:
    resp = await client.get(feed_url)

    if resp.status >= 400:
        raise RuntimeError(f"Bad status code {resp.status}")

    content = await resp.text()
    d = feedparser.parse(content)
    if d.bozo:
        raise RuntimeError("Feed is bad formatted")

    metadata = d["channel"]
    feed = Feed(
        url=feed_url,
        link=metadata.get("link"),
        title=metadata.get("title"),
        description=metadata.get("description", ""),
    )

    return feed


async def fetch_feed(
    client: aiohttp.ClientSession, feed: Feed
) -> tuple[list[Any], str]:
    headers = {"ETag": feed.etag} if feed.etag else {}
    resp = await client.get(feed.url, headers=headers)

    if resp.status == 304:
        return [], feed.etag

    if resp.status >= 400:
        raise RuntimeError(f"Bad status code {resp.status}")

    content = await resp.text()
    d = feedparser.parse(content)
    if d.bozo:
        raise RuntimeError("Feed is bad formatted")

    new_etag = resp.headers.get("ETag", "")
    return d.entries, new_etag


async def fetch_post(client: aiohttp.ClientSession, post_url: str) -> str:
    resp = await client.get(post_url)
    if resp.status >= 400:
        raise RuntimeError(f"Bad status code {resp.status}")

    return await resp.text()



from datetime import datetime, timezone
from typing import List
from sqlalchemy import ForeignKey, Boolean, Text
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    mapped_column,
    relationship,
)


class Base(DeclarativeBase):
    pass


class Feed(Base):
    __tablename__ = "feeds"

    id: Mapped[int] = mapped_column(primary_key=True)
    url: Mapped[str] = mapped_column(unique=True, index=True)
    link: Mapped[str] = mapped_column(nullable=True)
    title: Mapped[str]
    description: Mapped[str] = mapped_column(nullable=True)
    posts: Mapped[List["Post"]] = relationship(
        back_populates="feed",
        cascade="all, delete",
        passive_deletes=True,
    )

    etag: Mapped[str] = mapped_column(nullable=True)
    created_at: Mapped[datetime] = mapped_column(default=datetime.now(timezone.utc))
    last_updated_at: Mapped[datetime] = mapped_column(
        default=datetime.now(timezone.utc),
        onupdate=datetime.now(timezone.utc),
    )

    def __repr__(self) -> str:
        return f"Feed(id={self.id!r}, url={self.url!r}, title={self.title!r}, created_at={self.created_at!r}, last_updated_at={self.last_updated_at!r})"


class Post(Base):
    __tablename__ = "posts"

    id: Mapped[int] = mapped_column(primary_key=True)
    feed_id: Mapped[int] = mapped_column(ForeignKey("feeds.id"))
    feed: Mapped[Feed] = relationship(back_populates="posts")
    url: Mapped[str] = mapped_column(unique=True, index=True)
    author: Mapped[str] = mapped_column(nullable=True)
    title: Mapped[str]
    summary: Mapped[str] = mapped_column(nullable=True)
    content: Mapped[str] = mapped_column(Text(), nullable=True)

    read: Mapped[bool] = mapped_column(Boolean(), default=False)
    favorite: Mapped[bool] = mapped_column(Boolean(), default=False)
    saved_for_later: Mapped[bool] = mapped_column(Boolean(), default=False)

    published_at: Mapped[datetime] = mapped_column(
        default=datetime.now(timezone.utc),
        index=True,
    )
    last_updated_at: Mapped[datetime] = mapped_column(
        default=datetime.now(timezone.utc),
        onupdate=datetime.now(timezone.utc),
    )

    def __repr__(self) -> str:
        return f"Post(id={self.id!r}, url={self.url!r}, title={self.title!r}, favorite={self.favorite!r}, saved_for_later={self.saved_for_later!r}, published_at={self.published_at!r}, last_updated_at={self.last_updated_at!r})"



from textual.app import ComposeResult
from textual.containers import Horizontal, VerticalScroll
from textual.widgets import Static, Button
from textual.screen import ModalScreen


class ConfirmModal(ModalScreen[bool]):
    BINDINGS = [
        ("escape", "app.pop_screen", "Quit"),
        ("n", "cancel", "Cancel"),
        ("q", "app.pop_screen", "Quit"),
        ("y", "accept", "Accept"),
    ]

    def __init__(self, message: str, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.message = message

    def compose(self) -> ComposeResult:
        with VerticalScroll(classes="modal modal--help"):
            yield Static(self.message)
            yield Horizontal(
                Button(label="(N)o", id="no").focus(),
                Button(label="(Y)es", id="yes"),
                classes="modal__options",
            )

    def on_mount(self) -> None:
        self.modal = self.query_one(".modal")
        self.modal.border_title = "confirm"
        self.modal.border_subtitle = "n cancel · y accept · q quit"

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "no":
            self.dismiss(False)
        else:
            self.dismiss(True)

    def action_quit(self) -> None:
        self.dismiss(False)

    def action_cancel(self) -> None:
        self.dismiss(False)

    def action_accept(self) -> None:
        self.dismiss(True)



import asyncio
from datetime import datetime, timezone
from enum import Enum
import aiohttp
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from textual import on, work
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.reactive import reactive
from lazyfeed.confirm_modal import ConfirmModal
from lazyfeed.db import init_db
from lazyfeed.feeds import fetch_feed
from lazyfeed.help_modal import HelpModal
from lazyfeed.models import Post, Feed
from lazyfeed.repositories import FeedRepository, PostRepository
from lazyfeed.settings import Settings
from lazyfeed.tabloid import Tabloid

ActiveView = Enum("ActiveView", ["IDLE", "ALL", "PENDING", "SAVED", "FAV"])


class LazyFeedApp(App):
    """
    A Textual based application to read RSS feeds in
    the terminal.
    """

    TITLE = "lazyfeed"
    CSS_PATH = "global.tcss"
    ENABLE_COMMAND_PALETTE = False

    BINDINGS = [
        Binding("?", "display_help", "Display Help Message", show=False),
        Binding("q", "quit", "Quit", show=False),
        Binding("escape", "quit", "Quit", show=False),
        Binding("r", "refresh", "Reload", show=False),
    ]

    active_view: reactive[ActiveView] = reactive(ActiveView.IDLE)

    def __init__(self, session: Session, settings: Settings):
        self._session = session
        self.feeds_repository = FeedRepository(self._session)
        self.post_repository = PostRepository(self._session)
        self._settings = settings
        self._theme = settings.app.theme

        super().__init__()

    def compose(self) -> ComposeResult:
        yield Tabloid()

    def action_display_help(self) -> None:
        self.push_screen(HelpModal())

    def action_refresh(self) -> None:
        self.fetch_posts()

    async def action_quit(self) -> None:
        if self._settings.app.auto_mark_as_read:
            self.post_repository.mark_all_as_read()

        self.app.exit()

    async def on_mount(self) -> None:
        self.tabloid = self.query_one(Tabloid)
        self.fetch_posts()

    def get_css_variables(self) -> dict:
        css_variables = super().get_css_variables()
        if self._theme:
            color_system = self._theme.to_color_system().generate()
            return {**css_variables, **color_system}

        return css_variables

    @on(Tabloid.LoadAllPosts)
    async def set_view_to_all(self) -> None:
        self.active_view = ActiveView.ALL

    @on(Tabloid.LoadAllNewPosts)
    async def set_view_to_pending(self) -> None:
        self.active_view = ActiveView.PENDING

    @on(Tabloid.LoadSavedPosts)
    async def set_view_to_saved(self) -> None:
        self.active_view = ActiveView.SAVED

    @on(Tabloid.LoadFavPosts)
    async def set_view_to_fav(self) -> None:
        self.active_view = ActiveView.FAV

    @on(Tabloid.OpenPost)
    def open_item(self, message: Tabloid.OpenPost) -> None:
        post_in_db = self.post_repository.get(message.post_id)
        if not post_in_db:
            self.notify(
                "Unable to open the selected item",
                severity="error",
            )
            return

        self.open_url(post_in_db.url)
        self.post_repository.update(message.post_id, read=True)
        row_removed = self._pop_post(f"{post_in_db.id}")

        if not row_removed:
            self.tabloid.update_cell(
                f"{post_in_db.id}",
                "title",
                self._gen_row_content(post_in_db)[2],
            )

    @on(Tabloid.SavePost)
    def save_for_later(self, message: Tabloid.SavePost) -> None:
        post_in_db = self.post_repository.get(message.post_id)
        if not post_in_db:
            self.notify(
                "Unable to save for later the selected item",
                severity="error",
            )
            return

        self.post_repository.update(
            message.post_id, saved_for_later=not post_in_db.saved_for_later
        )

        saved = "\uf02e" if post_in_db.saved_for_later else ""
        self.tabloid.update_cell(f"{message.post_id}", "saved", saved)

    @on(Tabloid.MarkPostAsFav)
    def mark_as_fav(self, message: Tabloid.MarkPostAsFav) -> None:
        post_in_db = self.post_repository.get(message.post_id)
        if not post_in_db:
            self.notify(
                "Unable to mark as fav the selected item",
                severity="error",
            )
            return

        self.post_repository.update(message.post_id, favorite=not post_in_db.favorite)

        fav = "\uf005" if post_in_db.favorite else ""
        self.tabloid.update_cell(f"{post_in_db.id}", "fav", fav)

    @on(Tabloid.MarkPostAsRead)
    def mark_item_as_read(self, message: Tabloid.MarkPostAsRead) -> None:
        post_in_db = self.post_repository.get(message.post_id)
        if not post_in_db:
            self.notify(
                "Unable to mark as read the selected item",
                severity="error",
            )
            return

        self.post_repository.update(message.post_id, read=not post_in_db.read)
        row_removed = self._pop_post(f"{post_in_db.id}")
        if row_removed:
            return

        self.tabloid.update_cell(
            f"{post_in_db.id}",
            "title",
            self._gen_row_content(post_in_db)[2],
        )

    @on(Tabloid.MarkAllPostsAsRead)
    def mark_all_items_as_read(self) -> None:
        def check_confirmation(response: bool | None) -> None:
            if response:
                self._mark_all_post_as_read()

        if self._settings.app.ask_before_marking_as_read:
            self.push_screen(
                ConfirmModal("Are you sure that you want to mark all items as read?"),
                check_confirmation,
            )
        else:
            self._mark_all_post_as_read()

    def watch_active_view(self, old_view: ActiveView, new_view: ActiveView) -> None:
        if old_view == new_view:
            return

        if new_view == ActiveView.IDLE:
            self.tabloid.border_title = "lazyfeed"
            return

        if new_view == ActiveView.ALL:
            self.tabloid.border_title = "lazyfeed/all"
            self._load_posts()
            return

        if new_view == ActiveView.PENDING:
            self.tabloid.border_title = "lazyfeed"
            self._load_posts(read=False)
            return

        if new_view == ActiveView.SAVED:
            self.tabloid.border_title = "lazyfeed/saved"
            self._load_posts(saved_for_later=True)
            return

        if new_view == ActiveView.FAV:
            self.tabloid.border_title = "lazyfeed/fav"
            self._load_posts(favorite=True)

    def _gen_row_content(self, post: Post) -> tuple[str, str, str]:
        """
        Generate the content for a row for the given post.
        """

        saved = "\uf02e" if post.saved_for_later else ""
        fav = "\uf005" if post.favorite else ""
        label = f"[bold][{post.feed.title}][/bold] {post.title}"
        if post.read:
            label = f"[strike]{label}[/]"

        return saved, fav, label

    def _pop_post(self, row_id: str) -> bool:
        if self.active_view == ActiveView.PENDING and not self._settings.app.show_read:
            self.tabloid.remove_row(row_id)
            return True

        return False

    def _load_posts(self, **kwargs) -> None:
        self.tabloid.clear()

        sort_by = self._settings.app.sort_by
        sort_order = self._settings.app.sort_order
        sort_order_ascending = sort_order == "ascending" or sort_order == "asc"

        posts = self.post_repository.get_sorted(
            sort_by=sort_by,
            ascending=sort_order_ascending,
            **kwargs,
        )

        for post in posts:
            self.tabloid.add_row(*self._gen_row_content(post), key=f"{post.id}")

    def _mark_all_post_as_read(self) -> None:
        self.tabloid.loading = True

        try:
            self.post_repository.mark_all_as_read()
            self.active_view = ActiveView.IDLE

            self.tabloid.clear()

            self.notify(
                "All items have been marked as read",
                severity="information",
            )
        except Exception:
            self.notify(
                "Something went wrong while marking all items as read!",
                severity="error",
            )
        finally:
            self.tabloid.loading = False

    def _process_post(self, feed_id: int, entry: dict) -> Post | None:
        entry_link = entry.get("link")
        entry_title = entry.get("title")
        entry_summary = entry.get("summary")
        entry_published_parsed = entry.get("published_parsed")

        if not entry_link or not entry_title:
            return None

        published_at = None
        if entry_published_parsed:
            published_at = datetime(
                *entry_published_parsed[:6],
                tzinfo=timezone.utc,
            )

        return Post(
            feed_id=feed_id,
            url=entry_link,
            title=entry_title,
            summary=entry_summary,
            published_at=published_at,
        )

    async def _process_posts(self, client: aiohttp.ClientSession, feed: Feed) -> None:
        self.log(f"Processing posts from {feed.title}")

        try:
            result = await fetch_feed(client, feed)

            new_posts = []
            entries, etag = result
            if etag:
                self.feeds_repository.update(feed.id, etag=etag)

            for entry in entries:
                posts_in_db = self.post_repository.get_by_attributes(url=entry.link)
                if posts_in_db:
                    continue

                post = self._process_post(feed.id, entry)
                if post:
                    new_posts.append(post)

            self.post_repository.add_in_batch(new_posts)
        except Exception:
            self.notify(
                f"Something bad happened while fetching '{feed.title}'",
                severity="error",
            )

    @work(exclusive=True)
    async def fetch_posts(self) -> None:
        self.tabloid.loading = True

        feeds = self.feeds_repository.get_all()
        if not feeds:
            self.notify(
                "You need to add some feeds first!",
                severity="warning",
            )
            self.tabloid.loading = False
            return

        timeout = aiohttp.ClientTimeout(
            total=self._settings.client.timeout,
            connect=self._settings.client.connect_timeout,
        )
        headers = self._settings.client.headers
        async with aiohttp.ClientSession(timeout=timeout, headers=headers) as client:
            tasks = [self._process_posts(client, feed) for feed in feeds]
            await asyncio.gather(*tasks)

        self.active_view = ActiveView.PENDING
        self.tabloid.loading = False
        self.tabloid.focus()


if __name__ == "__main__":
    import os
    from lazyfeed.models import Feed

    os.environ["APP__DB_URL"] = "sqlite:///:memory:"

    settings = Settings()
    engine = create_engine(settings.app.db_url)
    init_db(engine)

    with Session(engine) as session:
        feed_repo = FeedRepository(session)
        feed_repo.add(
            Feed(
                title="Lorem RSS",
                url="https://lorem-rss.herokuapp.com/feed",
            )
        )

        app = LazyFeedApp(session, settings)
        app.run()
